import logging
import os
from typing import BinaryIO

import boto3
from botocore.client import Config
from botocore.exceptions import ClientError
from dotenv import load_dotenv

load_dotenv()

log = logging.getLogger("udata-hydra")

MINIO_FOLDER = os.environ.get("MINIO_FOLDER", "folder")


def get_resource_minio_url(key: str, resource_id: str) -> str:
    """Returns location of given resource in minio once it is saved"""
    return (
        os.getenv("MINIO_URL")
        + "/"
        + os.getenv("MINIO_BUCKET")
        + "/"
        + MINIO_FOLDER
        + "/"
        + key
        + "/"
        + resource_id
    )


def save_resource_to_minio(resource_file: BinaryIO, key: str, resource_id: str) -> None:
    log.info("Saving to minio")
    log.debug(f'Bucket is {os.getenv("MINIO_BUCKET")}')
    s3 = boto3.client(
        "s3",
        endpoint_url=os.getenv("MINIO_URL"),
        aws_access_key_id=os.getenv("MINIO_USER"),
        aws_secret_access_key=os.getenv("MINIO_PWD"),
        config=Config(signature_version="s3v4"),
    )
    try:
        with open(resource_file.name, "rb") as f:
            s3.upload_fileobj(
                f,
                os.getenv("MINIO_BUCKET"),
                MINIO_FOLDER + "/" + key + "/" + resource_id,
            )
        log.info(
            f"Resource saved into minio at {get_resource_minio_url(key, resource_id)}"
        )
    except ClientError as e:
        log.error(e)


def delete_resource_from_minio(key: str, resource_id: str) -> None:
    log.info("Deleting from minio")
    s3 = boto3.client(
        "s3",
        endpoint_url=os.getenv("MINIO_URL"),
        aws_access_key_id=os.getenv("MINIO_USER"),
        aws_secret_access_key=os.getenv("MINIO_PWD"),
        config=Config(signature_version="s3v4"),
    )
    try:
        s3.delete_object(
            Bucket=os.getenv("MINIO_BUCKET"),
            Key=MINIO_FOLDER + "/" + key + "/" + resource_id,
        )
        log.info(
            f"Resource deleted from minio at {get_resource_minio_url(key, resource_id)}"
        )
    except ClientError as e:
        log.error(e)



import os
import logging

from unittest.mock import MagicMock

import asyncpg

log = logging.getLogger("udata-hydra")
context = {}


def monitor():
    if "monitor" in context:
        return context["monitor"]
    monitor = MagicMock()
    monitor.set_status = lambda x: log.debug(x)
    monitor.init = lambda **kwargs: log.debug(
        f"Starting udata-hydra... {kwargs}"
    )
    context["monitor"] = monitor
    return context["monitor"]


async def pool():
    if "pool" not in context:
        dsn = os.getenv(
            "DATABASE_URL",
            "postgres://postgres:postgres@localhost:5432/postgres",
        )
        context["pool"] = await asyncpg.create_pool(dsn=dsn, max_size=50)
    return context["pool"]



# -- crawler settings -- #

# sql LIKE syntax
EXCLUDED_PATTERNS = [
    'http%geo.data.gouv.fr%',
    # opendatasoft shp
    '%?format=shp%',
]
# no backoff for those domains
NO_BACKOFF_DOMAINS = [
    'static.data.gouv.fr',
    'www.data.gouv.fr',
    # dead domain, no need to backoff
    'inspire.data.gouv.fr',
]
# max number of _completed_ requests per domain per period
BACKOFF_NB_REQ = 180
BACKOFF_PERIOD = 360  # in seconds
# crawl batch size, beware of open file limits
BATCH_SIZE = 100
# crawl url if last check is older than
SINCE = '1w'
# seconds to wait for between batches
SLEEP_BETWEEN_BATCHES = 60

# -- Webhook integration config -- #
ENABLE_WEBHOOK = True
UDATA_URI = 'http://dev.local:7000/api/2'
UDATA_URI_API_KEY = ''

# Save resource to minio or not
SAVE_TO_MINIO = False



import json
import logging
import sys
import time

from collections import defaultdict
from datetime import datetime, timedelta
from requests.exceptions import SSLError
from urllib.parse import urlparse

import aiohttp
import asyncio

from humanfriendly import parse_timespan

from udata_hydra import config, context
from udata_hydra.datalake_service import process_resource
from udata_hydra.utils.http import send

log = logging.getLogger("udata-hydra")

results = defaultdict(int)

STATUS_OK = "ok"
STATUS_TIMEOUT = "timeout"
STATUS_ERROR = "error"


async def insert_check(data: dict):
    if "headers" in data:
        data["headers"] = json.dumps(data["headers"])
    columns = ",".join(data.keys())
    # $1, $2...
    placeholders = ",".join([f"${x + 1}" for x in range(len(data.values()))])
    q = f"""
        INSERT INTO checks ({columns})
        VALUES ({placeholders})
        RETURNING id
    """
    pool = await context.pool()
    async with pool.acquire() as connection:
        last_check = await connection.fetchrow(q, *data.values())
        q = """UPDATE catalog SET last_check = $1 WHERE url = $2"""
        await connection.execute(q, last_check["id"], data["url"])
    return last_check["id"]


async def update_check_and_catalog(check_data: dict) -> None:
    """Update the catalog and checks tables"""
    context.monitor().set_status("Updating checks and catalog...")
    check_data['resource_id'] = str(check_data['resource_id'])

    pool = await context.pool()
    async with pool.acquire() as connection:
        q = f"""
            SELECT * FROM catalog JOIN checks
            ON catalog.last_check = checks.id
            WHERE catalog.resource_id = '{check_data['resource_id']}';
        """
        last_checks = await connection.fetch(q)

        if len(last_checks) == 0:
            # In case we are doing our first check for given URL
            rows = await connection.fetch(
                f"""
                SELECT resource_id, dataset_id, priority, initialization
                FROM catalog
                WHERE resource_id = '{check_data['resource_id']}';
            """
            )
            last_checks = [
                {
                    "resource_id": row[0],
                    "dataset_id": row[1],
                    "priority": row[2],
                    "initialization": row[3],
                    "status": None,
                    "timeout": None,
                }
                for row in rows
            ]

        # There could be multiple resources pointing to the same URL
        for last_check in last_checks:
            if config.ENABLE_WEBHOOK:
                is_first_check = last_check["status"] is None
                status_has_changed = (
                    "status" in check_data
                    and check_data["status"] != last_check["status"]
                )
                status_no_longer_available = (
                    "status" not in check_data
                    and last_check["status"] is not None
                )
                timeout_has_changed = (
                    check_data["timeout"] != last_check["timeout"]
                )

                if (
                    is_first_check
                    or status_has_changed
                    or status_no_longer_available
                    or timeout_has_changed
                ):
                    log.debug("Sending message to udata...")
                    document = dict()
                    document['check:status'] = check_data['status'] if status_has_changed else last_check["status"]
                    document['check:timeout'] = check_data['timeout']
                    document['check:check_date'] = str(datetime.now())
                    await send(dataset_id=last_check["dataset_id"],
                               resource_id=last_check["resource_id"],
                               document=document)

        log.debug("Updating priority...")
        await connection.execute(
            f"""
            UPDATE catalog SET priority = FALSE, initialization = FALSE
            WHERE resource_id = '{check_data['resource_id']}';
        """
        )

    await insert_check(check_data)


async def is_backoff(domain):
    no_backoff = [f"'{d}'" for d in config.NO_BACKOFF_DOMAINS]
    no_backoff = f"({','.join(no_backoff)})"
    since = datetime.utcnow() - timedelta(seconds=config.BACKOFF_PERIOD)
    pool = await context.pool()
    async with pool.acquire() as connection:
        res = await connection.fetchrow(
            f"""
            SELECT COUNT(*) FROM checks
            WHERE domain = $1
            AND created_at >= $2
            AND domain NOT IN {no_backoff}
        """,
            domain,
            since,
        )
        return res["count"] >= config.BACKOFF_NB_REQ, res["count"]


def fix_surrogates(value):
    """FIX Unicode low surrogate must follow a high surrogate.
    eg in 'TREMI_2017-R\xe9sultats enqu\xeate bruts.csv'
    """
    return value.encode("utf-8", "surrogateescape").decode("utf-8", "replace")


def convert_headers(headers):
    """
    Convert headers from CIMultiDict to dict

    :warning: this will only take the first value for a given header
    key but multidict is not json serializable
    """
    if not headers:
        return {}
    _headers = {}
    for k in headers.keys():
        value = fix_surrogates(headers[k])
        _headers[k.lower()] = value
    return _headers


async def check_url(row, session, sleep=0, method="get"):
    log.debug(f"check {row}, sleep {sleep}")

    if sleep:
        await asyncio.sleep(sleep)

    url_parsed = urlparse(row["url"])
    domain = url_parsed.netloc
    if not domain:
        log.warning(f"[warning] not netloc in url, skipping {row['url']}")
        await update_check_and_catalog(
            {
                "resource_id": row["resource_id"],
                "url": row["url"],
                "error": "Not netloc in url",
                "timeout": False,
            }
        )
        return STATUS_ERROR

    should_backoff, nb_req = await is_backoff(domain)
    if should_backoff:
        log.info(f"backoff {domain} ({nb_req})")
        context.monitor().add_backoff(domain, nb_req)
        # TODO: maybe just skip this url, it should come back in the next batch anyway
        # but won't it accumulate too many backoffs in the end? is this a problem?
        return await check_url(
            row, session, sleep=config.BACKOFF_PERIOD / config.BACKOFF_NB_REQ
        )
    else:
        context.monitor().remove_backoff(domain)

    try:
        start = time.time()
        timeout = aiohttp.ClientTimeout(total=5)
        _method = getattr(session, method)
        async with _method(
            row["url"], timeout=timeout, allow_redirects=True
        ) as resp:
            end = time.time()
            resp.raise_for_status()

            # Download resource, store on Minio if CSV and produce resource.analysed message
            res = await process_resource(row["url"], row["dataset_id"], str(row["resource_id"]), resp)

            await update_check_and_catalog(
                {
                    "resource_id": row["resource_id"],
                    "url": row["url"],
                    "domain": domain,
                    "status": resp.status,
                    "headers": convert_headers(resp.headers),
                    "timeout": False,
                    "response_time": end - start,
                    "error": res["error"],
                    "checksum": res["checksum"],
                    "filesize": res["filesize"],
                    "mime_type": res["mime_type"]
                }
            )
            return STATUS_OK
    except asyncio.exceptions.TimeoutError:
        await update_check_and_catalog(
            {
                "resource_id": row["resource_id"],
                "url": row["url"],
                "domain": domain,
                "timeout": True,
            }
        )
        return STATUS_TIMEOUT
    # TODO: debug AssertionError, should be caught in DB now
    # File "[...]aiohttp/connector.py", line 991, in _create_direct_connection
    # assert port is not None
    # UnicodeError: encoding with 'idna' codec failed (UnicodeError: label too long)
    # eg http://%20Localisation%20des%20acc%C3%A8s%20des%20offices%20de%20tourisme
    except (
        aiohttp.client_exceptions.ClientError,
        AssertionError,
        UnicodeError,
        SSLError,
    ) as e:
        error = getattr(e, "message", None) or str(e)
        await update_check_and_catalog(
            {
                "resource_id": row["resource_id"],
                "url": row["url"],
                "domain": domain,
                "timeout": False,
                "error": fix_surrogates(error),
                "headers": convert_headers(getattr(e, "headers", {})),
                "status": getattr(e, "status", None),
            }
        )
        log.error(f"{row['url']}, {e}")
        return STATUS_ERROR


async def crawl_urls(to_parse):
    context.monitor().set_status("Crawling urls...")
    tasks = []
    async with aiohttp.ClientSession(timeout=None) as session:
        for row in to_parse:
            tasks.append(check_url(row, session))
        for task in asyncio.as_completed(tasks):
            result = await task
            results[result] += 1
            context.monitor().refresh(results)


def get_excluded_clause():
    return " AND ".join(
        [f"catalog.url NOT LIKE '{p}'" for p in config.EXCLUDED_PATTERNS]
    )


async def crawl_batch():
    """Crawl a batch from the catalog"""
    context.monitor().set_status("Getting a batch from catalog...")
    pool = await context.pool()
    async with pool.acquire() as connection:
        excluded = get_excluded_clause()
        # first urls that are prioritised
        q = f"""
            SELECT * FROM (
                SELECT catalog.url, dataset_id, resource_id
                FROM catalog
                WHERE {excluded}
                AND deleted = False
                AND priority = True
            ) s
            ORDER BY random() LIMIT {config.BATCH_SIZE};
        """
        to_check = await connection.fetch(q)
        # then urls without checks
        if len(to_check) < config.BATCH_SIZE:
            q = f"""
                SELECT * FROM (
                    SELECT catalog.url, dataset_id, resource_id
                    FROM catalog
                    WHERE catalog.last_check IS NULL
                    AND {excluded}
                    AND deleted = False
                    AND priority = False
                ) s
                ORDER BY random() LIMIT {config.BATCH_SIZE};
            """
            to_check += await connection.fetch(q)
        # if not enough for our batch size, handle outdated checks
        if len(to_check) < config.BATCH_SIZE:
            since = parse_timespan(config.SINCE)  # in seconds
            since = datetime.utcnow() - timedelta(seconds=since)
            limit = config.BATCH_SIZE - len(to_check)
            q = f"""
            SELECT * FROM (
                SELECT catalog.url, dataset_id, catalog.resource_id
                FROM catalog, checks
                WHERE catalog.last_check IS NOT NULL
                AND {excluded}
                AND catalog.last_check = checks.id
                AND checks.created_at <= $1
                AND catalog.deleted = False
                AND catalog.priority = False
            ) s
            ORDER BY random() LIMIT {limit};
            """
            to_check += await connection.fetch(q, since)

    if len(to_check):
        await crawl_urls(to_check)
    else:
        context.monitor().set_status("Nothing to crawl for now.")
        await asyncio.sleep(config.SLEEP_BETWEEN_BATCHES)


async def crawl(iterations=-1):
    """Launch crawl batches

    :iterations: for testing purposes (break infinite loop)
    """
    try:
        context.monitor().init(
            SINCE=config.SINCE,
            BATCH_SIZE=config.BATCH_SIZE,
            BACKOFF_NB_REQ=config.BACKOFF_NB_REQ,
            BACKOFF_PERIOD=config.BACKOFF_PERIOD,
        )
        while iterations != 0:
            await crawl_batch()
            iterations -= 1
    finally:
        pool = await context.pool()
        await pool.close()


def setup_logging():
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter("%(asctime)s %(levelname)s %(message)s")
    handler.setFormatter(formatter)
    log.addHandler(handler)
    log.setLevel(logging.DEBUG)


def run():
    """Main function

    :iterations: for testing purposes (break infinite loop)
    """
    setup_logging()
    try:
        asyncio.get_event_loop().run_until_complete(crawl())
    except KeyboardInterrupt:
        pass
    finally:
        context.monitor().teardown()


if __name__ == "__main__":
    run()



import logging
import csv
import os

from pathlib import Path
from tempfile import NamedTemporaryFile

import aiohttp
import asyncpg
from humanfriendly import parse_size
from minicli import cli, run, wrap
from progressist import ProgressBar

log = logging.getLogger("udata-hydra")
log.setLevel(os.getenv("LOGLEVEL", logging.INFO))
logging.basicConfig()

CATALOG_URL = 'https://www.data.gouv.fr/fr/datasets/r/4babf5f2-6a9c-45b5-9144-ca5eae6a7a6d'

context = {}


@cli
async def init_db(drop=False, table=None, index=False, reindex=False):
    """Create the DB structure"""
    log.info("Initializing database...")
    if drop:
        if table == "catalog" or not table:
            await context["conn"].execute("DROP TABLE IF EXISTS catalog")
        if table == "checks" or not table:
            await context["conn"].execute("DROP TABLE IF EXISTS checks")
    await context["conn"].execute(
        """
        CREATE TABLE IF NOT EXISTS catalog(
            id serial PRIMARY KEY,
            dataset_id VARCHAR(24),
            resource_id UUID,
            url VARCHAR,
            deleted BOOLEAN NOT NULL,
            last_check INT,
            priority BOOLEAN NOT NULL,
            initialization BOOLEAN NOT NULL DEFAULT FALSE,
            UNIQUE(dataset_id, resource_id, url)
        )
    """
    )
    await context["conn"].execute(
        """
        CREATE TABLE IF NOT EXISTS checks(
            id serial PRIMARY KEY,
            resource_id UUID,
            url VARCHAR,
            domain VARCHAR,
            created_at TIMESTAMP DEFAULT NOW(),
            status INT,
            headers JSONB,
            timeout BOOLEAN NOT NULL,
            response_time FLOAT,
            error VARCHAR,
            checksum VARCHAR,
            filesize BIGINT,
            mime_type VARCHAR
        )
    """
    )
    if reindex:
        await context["conn"].execute(
            """
            DROP INDEX IF EXISTS url_idx;
            DROP INDEX IF EXISTS domain_idx;
        """
        )
    if index or reindex:
        await context["conn"].execute(
            """
            CREATE INDEX IF NOT EXISTS url_idx ON checks (url);
            CREATE INDEX IF NOT EXISTS domain_idx ON checks (domain);
        """
        )


async def download_file(url, fd):
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as resp:
            while True:
                chunk = await resp.content.read(1024)
                if not chunk:
                    break
                fd.write(chunk)


@cli
async def load_catalog(url=CATALOG_URL):
    """Load the catalog into DB from CSV file"""
    try:
        log.info(f"Downloading catalog from {url}...")
        with NamedTemporaryFile(delete=False) as fd:
            await download_file(url, fd)
        log.info("Upserting catalog in database...")
        # consider everything deleted, deleted will be updated when loading new catalog
        await context["conn"].execute("UPDATE catalog SET deleted = TRUE")
        with open(fd.name) as fd:
            reader = csv.DictReader(fd, delimiter=";")
            rows = list(reader)
            bar = ProgressBar(total=len(rows))
            for row in bar.iter(rows):
                await context["conn"].execute(
                    """
                    INSERT INTO catalog (dataset_id, resource_id, url, deleted, priority, initialization)
                    VALUES ($1, $2, $3, FALSE, FALSE, TRUE)
                    ON CONFLICT (dataset_id, resource_id, url) DO UPDATE SET deleted = FALSE
                """,
                    row["dataset.id"],
                    row["id"],
                    row["url"],
                )
        log.info("Catalog successfully upserted into DB.")
    except Exception as e:
        raise e
    finally:
        fd.close()
        os.unlink(fd.name)


@cli
async def csv_sample(size=1000, download=False, max_size="100M"):
    """Get a csv sample from latest checks

    :size: Size of the sample (how many files to query)
    :download: Download files or just list them
    :max_size: Maximum size for one file (from headers)
    """
    max_size = parse_size(max_size)
    start_q = f"""
        SELECT catalog.resource_id, catalog.dataset_id, checks.url,
            checks.headers->>'content-type' as content_type,
            checks.headers->>'content-length' as content_length
        FROM checks, catalog
        WHERE catalog.last_check = checks.id
        AND checks.headers->>'content-type' LIKE '%csv%'
        AND checks.status >= 200 and checks.status < 400
        AND CAST(checks.headers->>'content-length' AS INTEGER) <= {max_size}
    """
    end_q = f"""
        ORDER BY RANDOM()
        LIMIT {size / 2}
    """
    # get remote stuff for half the sample
    q = f"""{start_q}
        -- ignore ODS, they're correctly formated from a datastore
        AND checks.url NOT LIKE '%/explore/dataset/%'
        AND checks.url NOT LIKE '%/api/datasets/1.0/%'
        -- ignore ours
        AND checks.domain <> 'static.data.gouv.fr'
        {end_q}
    """
    res = await context["conn"].fetch(q)
    # and from us for the rest
    q = f"""{start_q}
        AND checks.domain = 'static.data.gouv.fr'
        {end_q}
    """
    res += await context["conn"].fetch(q)

    data_path = Path("./data")
    dl_path = data_path / "downloaded"
    dl_path.mkdir(exist_ok=True, parents=True)
    if download:
        log.debug("Cleaning up...")
        (data_path / "_index.csv").unlink(missing_ok=True)
        [p.unlink() for p in Path(dl_path).glob("*.csv")]

    lines = []
    bar = ProgressBar(total=len(res))
    for r in bar.iter(res):
        line = dict(r)
        line["resource_id"] = str(line["resource_id"])
        filename = dl_path / f"{r['dataset_id']}_{r['resource_id']}.csv"
        line["filename"] = filename.__str__()
        lines.append(line)
        if not download:
            continue
        await download_file(r["url"], filename.open("wb"))
        with os.popen(f"file {filename} -b --mime-type") as proc:
            line["magic_mime"] = proc.read().lower().strip()
        line["real_size"] = filename.stat().st_size

    with (data_path / "_index.csv").open("w") as ofile:
        writer = csv.DictWriter(ofile, fieldnames=lines[0].keys())
        writer.writeheader()
        writer.writerows(lines)


@wrap
async def cli_wrapper():
    dsn = os.getenv(
        "DATABASE_URL", "postgres://postgres:postgres@localhost:5432/postgres"
    )
    context["conn"] = await asyncpg.connect(dsn=dsn)
    context["dsn"] = dsn
    yield
    await context["conn"].close()


if __name__ == "__main__":
    run()



from datetime import datetime
import hashlib
import logging
import os
import tempfile
from typing import BinaryIO

import magic
import pandas as pd

from aiohttp import ClientResponse
from dotenv import load_dotenv

from udata_hydra import config, context
from udata_hydra.utils.http import send
from udata_hydra.utils.json import is_json_file
from udata_hydra.utils.minio import save_resource_to_minio
from udata_hydra.utils.csv import detect_encoding, find_delimiter

load_dotenv()

log = logging.getLogger("udata-hydra")

MINIO_FOLDER = os.environ.get("MINIO_FOLDER", "folder")
MAX_FILESIZE_ALLOWED = os.environ.get("MAX_FILESIZE_ALLOWED", 1000)


async def download_resource(url: str, response: ClientResponse) -> BinaryIO:
    """
    Attempts downloading a resource from a given url.
    Returns the downloaded file object.
    Raises IOError if the resource is too large.
    """
    tmp_file = tempfile.NamedTemporaryFile(delete=False)

    if float(response.headers.get("content-length", -1)) > float(
        MAX_FILESIZE_ALLOWED
    ):
        raise IOError("File too large to download")

    chunk_size = 1024
    i = 0
    async for chunk in response.content.iter_chunked(chunk_size):
        if i * chunk_size < float(MAX_FILESIZE_ALLOWED):
            tmp_file.write(chunk)
        else:
            tmp_file.close()
            log.error(f"File {url} is too big, skipping")
            raise IOError("File too large to download")
        i += 1
    tmp_file.close()
    return tmp_file


async def process_resource(url: str, dataset_id: str, resource_id: str, response: ClientResponse) -> None:
    log.debug(
        "Processing task for resource {} in dataset {}".format(
            resource_id, dataset_id
        )
    )

    tmp_file = None
    mime_type = None
    sha1 = None
    filesize = None
    error = None

    try:
        tmp_file = await download_resource(url, response)

        # Get file size
        filesize = os.path.getsize(tmp_file.name)

        # Get checksum
        sha1 = await compute_checksum_from_file(tmp_file.name)

        # Check resource MIME type
        mime_type = magic.from_file(tmp_file.name, mime=True)
        if mime_type in ["text/plain", "text/csv", "application/csv"] and not is_json_file(tmp_file.name):
            # Save resource only if CSV
            try:
                # Try to detect encoding from suspected csv file. If fail, set up to utf8 (most common)
                with open(tmp_file.name, mode="rb") as f:
                    try:
                        encoding = detect_encoding(f)
                    # FIXME: catch exception more precisely
                    except Exception:
                        encoding = "utf-8"
                # Try to detect delimiter from suspected csv file. If fail, set up to None
                # (pandas will use python engine and try to guess separator itself)
                try:
                    delimiter = find_delimiter(tmp_file.name)
                # FIXME: catch exception more precisely
                except Exception:
                    delimiter = None

                # Try to read first 1000 rows with pandas
                pd.read_csv(tmp_file.name, sep=delimiter, encoding=encoding, nrows=1000)

                if config.SAVE_TO_MINIO:
                    save_resource_to_minio(tmp_file, dataset_id, resource_id)
                    storage_location = "/".join([
                        os.getenv("MINIO_URL"),
                        os.getenv("MINIO_BUCKET"),
                        MINIO_FOLDER,
                        dataset_id,
                        resource_id
                    ])
                    log.debug(
                        f"Sending message to udata for resource stored {resource_id} in dataset {dataset_id}"
                    )
                    document = {"store:data_location": storage_location}
                    await send(dataset_id=dataset_id,
                               resource_id=resource_id,
                               document=document)
            except ValueError:
                log.debug(
                    f"Resource {resource_id} in dataset {dataset_id} is not a CSV"
                )

        # Send udata a message for both CSV and non CSV resources
        log.debug(
            f"Sending a message to udata for resource analysed {resource_id} in dataset {dataset_id}"
        )
        document = {
            "analysis:error": None,
            "analysis:filesize": filesize,
            "analysis:mime": mime_type,
        }
        # Check if checksum has been modified
        # TODO: improve file modification logic
        checksum_modified = await has_checksum_been_modified(resource_id, sha1)
        if checksum_modified:
            document["analysis:checksum_last_modified"] = datetime.utcnow().isoformat()

        await send(dataset_id=dataset_id,
                   resource_id=resource_id,
                   document=document)
    except IOError:
        error = "File too large to download"
        document = {
            "analysis:error": error,
            "analysis:filesize": None,
            "analysis:mime": None,
        }
        await send(dataset_id=dataset_id,
                   resource_id=resource_id,
                   document=document)
    finally:
        if tmp_file:
            os.remove(tmp_file.name)
        return {
            "checksum": sha1,
            "error": error,
            "filesize": filesize,
            "mime_type": mime_type
        }


async def has_checksum_been_modified(resource_id, new_checksum):
    q = """
    SELECT
        checksum
    FROM checks
    WHERE resource_id = $1 and checksum IS NOT NULL
    ORDER BY created_at DESC
    LIMIT 1
    """
    pool = await context.pool()
    async with pool.acquire() as connection:
        data = await connection.fetch(q, resource_id)
        if data:
            if data[0]["checksum"] != new_checksum:
                return True
            else:
                return False
        else:
            # First check, thus we don't consider the checksum has been modified
            return False


async def compute_checksum_from_file(filename):
    # Compute sha1 in blocks
    sha1sum = hashlib.sha1()
    with open(filename, "rb") as f:
        block = f.read(2**16)
        while len(block) != 0:
            sha1sum.update(block)
            block = f.read(2**16)
    return sha1sum.hexdigest()



import csv
from cchardet import UniversalDetector


def find_delimiter(filename):
    """Find delimiter for eventual csv file"""
    sniffer = csv.Sniffer()
    with open(filename) as fp:
        delimiter = sniffer.sniff(fp.read(5000), delimiters=';,|\t').delimiter
    return delimiter


def detect_encoding(the_file):
    '''Detects file encoding using chardet based on N first lines
    '''
    detector = UniversalDetector()
    for line in the_file.readlines():
        detector.feed(line)
        if detector.done:
            break
    detector.close()
    return detector.result['encoding']



import aiohttp
import logging

from udata_hydra import config

log = logging.getLogger("udata-hydra")


async def send(dataset_id: str, resource_id: str, document: dict) -> None:
    if not config.UDATA_URI or not config.UDATA_URI_API_KEY:
        log.error("Missing udata URI and API key to send http query")
        return
    uri = f"{config.UDATA_URI}/datasets/{dataset_id}/resources/{resource_id}/extras/"
    headers = {"content-type": "application/json", "X-API-KEY": config.UDATA_URI_API_KEY}

    # Extras in udata can't be None
    document = {key: document[key] for key in document if document[key] is not None}

    async with aiohttp.ClientSession() as session:
        async with session.put(uri, json=document, headers=headers) as resp:
            body = await resp.text()
            if not resp.status == 200:
                log.error(f"udata reponded with a {resp.status} and content: {body}")
